from typing import TYPE_CHECKING, List
from datetime import datetime
from sqlalchemy_json import NestedMutableJson
from sqlalchemy.orm import declarative_mixin, declared_attr, object_session
import sqlalchemy as sa
from .helpers import get_jembewf


if TYPE_CHECKING:
    import jembewf

__all__ = ("ProcessMixin", "CantStartProcess")


class CantStartProcess(Exception):
    """Exception raised when process can't be started"""


@declarative_mixin
class ProcessMixin:
    """Mixin to be applied to Process SqlAlchemy model

    Process model keep track of all instances of flows (current and completed)
    it global variables and states.

    ProcessMixin should be applied class extended from
    flask_sqlalchemy.SqlAlchemy().Model who defines model in database
    """

    __tablename__ = "jwf_processes"

    __step_table_name__: str = "jwf_steps"
    __step_class_name__: str = "Step"

    id = sa.Column(sa.Integer, primary_key=True)

    flow_name = sa.Column(sa.String(250), nullable=False)

    is_running = sa.Column(
        sa.Boolean(), nullable=False, default=True, server_default=sa.true()
    )

    # process variables
    variables = sa.Column(NestedMutableJson)

    started_at = sa.Column(sa.DateTime, default=datetime.utcnow, nullable=False)
    ended_at = sa.Column(sa.DateTime)

    @declared_attr
    def steps(cls):
        return sa.orm.relationship(
            cls.get_step_class_name(),
            foreign_keys=f"{cls.get_step_class_name()}.process_id",
            back_populates="process",
            order_by=f"{cls.get_step_class_name()}.id",
        )

    def current_steps(self) -> List["jembewf.StepMisin"]:
        """Returns current active process steps"""
        session = object_session(self)
        step = get_jembewf().step_model
        return list(session.query(step).filter(step.is_active == True).all())

    def last_steps(self) -> List["jembewf.StepMisin"]:
        """Returns last steps of the process"""
        session = object_session(self)
        step = get_jembewf().step_model
        return list(session.query(step).filter(step.is_last_step == True).all())

    @property
    def flow(self) -> "jembewf.Flow":
        """Flow of the process instance"""
        return get_jembewf().flows[self.flow_name]

    @property
    def callback(self) -> "jembewf.FlowCallback":
        """FlowCallback of the process instance"""
        return self.flow.callback(self)

    @classmethod
    def can_start(cls, flow_name: str, **process_vars) -> bool:
        """Checks if process can be started

        Args:
            flow_name (str): name of the flow for witch we are doing check
            process_vars: global process variables

        Raises:
            ValueError: Can't start flow because it does not exist!

        Returns:
            bool: True if process can be started
        """
        process = cls._create_process(flow_name, **process_vars)
        return process.callback.can_start()

    @classmethod
    def create(cls, flow_name: str, **process_vars) -> "jembewf.ProcessMixin":
        """Creates process instance for flow named flow_name

        Args:
            flow_name (str): name of the flow for witch we are creating the process
            process_vars: global process variables

        Raises:
            ValueError: Can't create process because flow does not exist!

        Returns:
            jembewf.ProcessMixin: Instance of the process model
        """
        jwf = get_jembewf()
        process = cls._create_process(flow_name, **process_vars)
        if process.callback.can_start():
            # save process to db
            jwf.db.session.add(process)
            jwf.db.session.commit()

            process.callback.callback()

            # create steps for starting tasks
            for task_name in process.flow.starts_with_tasks:
                task = process.flow.tasks[task_name]
                jwf.step_model.create(process, task)
        else:
            raise CantStartProcess(
                f"Can't start process '{flow_name}' with process vars: {process_vars}"
            )
        return process

    def proceed(self) -> bool:
        """Proceed with process execution

        Check if any of active steps can transit to the next tasks, if so
        do the transition.

        Returns True if process proceed to new steps
        """
        proceded = False
        for step in self.current_steps():
            if not step.is_last_step:
                proceded = proceded or step.proceed()
        return proceded

    def check_is_running(self):
        """Check if process is still running and update is_running if necessary"""
        session = object_session(self)
        step = get_jembewf().step_model
        is_running = session.query(
            session.query(step).filter(step.is_active == True).exists()
        ).scalar()
        if self.is_running != is_running:
            self.is_running = is_running
            session.add(self)
            session.commit()
        return is_running

    @classmethod
    def _create_process(cls, flow_name: str, **process_vars) -> "jembewf.ProcessMixin":
        jwf = get_jembewf()
        try:
            flow = jwf.flows[flow_name]
        except KeyError as err:
            raise ValueError(
                f"Can't start flow '{flow_name}' because it does not exist!"
            ) from err

        process = jwf.process_model()
        process.flow_name = flow.name
        process.variables = process_vars

        return process

    @classmethod
    def get_step_table_name(cls) -> str:
        """Returns name of Process table defined in cls.__step_table_name__

        It's used to create relationship between steps and steps.
        """
        try:
            return cls.__step_table_name__
        except AttributeError as err:
            raise AttributeError(
                f"Attribute __step_table_name__ for '{cls.__name__}' is not defined"
            ) from err

    @classmethod
    def get_step_class_name(cls) -> str:
        """Returns name of Process class defined in cls.__step_class_name__

        It's used to create relationship between steps and steps.
        """
        try:
            return cls.__step_class_name__
        except AttributeError as err:
            raise AttributeError(
                f"Attribute __step_class_name__ for '{cls.__name__}' is not defined"
            ) from err

    def __repr__(self):
        return f"<Process #{self.id}: '{self.flow_name}'>"



from typing import TYPE_CHECKING, Dict, Optional, Type
import json
from .flow import Flow, FlowCallback
from .task import Task, TaskCallback
from .transition import Transition, TransitionCallback
from .process_mixin import ProcessMixin, CantStartProcess
from .step_mixin import StepMixin
from .helpers import get_jembewf


if TYPE_CHECKING:
    from flask import Flask
    from flask_sqlalchemy import SQLAlchemy
    import jembewf


__all__ = (
    "get_jembewf",
    "JembeWF",
    "Flow",
    "FlowCallback",
    "Task",
    "TaskCallback",
    "Transition",
    "TransitionCallback",
    "ProcessMixin",
    "StepMixin",
    "CantStartProcess",
)


class JembeWF:
    """Jembe Workflow Management extension for Flask framework"""

    def __init__(
        self,
        app: Optional["Flask"] = None,
        db: Optional["SQLAlchemy"] = None,
        process_model: Optional[Type["jembewf.ProcessMixin"]] = None,
        step_model: Optional[Type["jembewf.StepMixin"]] = None,
    ) -> None:

        self.flows: Dict[str, "jembewf.Flow"] = {}
        self.initialised = False

        self.db: "SQLAlchemy"
        if db is not None:
            self.db = db

        self.process_model: Type["jembewf.ProcessMixin"]
        if process_model is not None:
            self.process_model = process_model

        self.step_model: Type["jembewf.StepMixin"]
        if step_model is not None:
            self.step_model = step_model

        if app is not None:
            if process_model is None or step_model is None or db is None:
                raise Exception(
                    "When you provide app parameter then "
                    "db, process_model and step_model must also be provided in other to initialise JembeWf."
                )
            self.init_app(app, self.db, self.process_model, self.step_model)

    def init_app(
        self,
        app: "Flask",
        db: Optional["SQLAlchemy"] = None,
        process_model: Optional[Type["jembewf.ProcessMixin"]] = None,
        step_model: Optional[Type["jembewf.StepMixin"]] = None,
    ) -> None:
        """Initialize JembeWF Flask extension

        Args:
            app (Flask): Flask instance
        """
        if "jembewf" in app.extensions:
            raise Exception("JembeWF is already registred at Flask app")
        # check if db is provided
        if db is None and not hasattr(self, "db"):
            raise Exception("JembeWF 'db' must be provided in __init__ or in init_app")
        if db is not None:
            self.db = db

        # check if process_model is provided
        if process_model is None and not hasattr(self, "process_model"):
            raise Exception(
                "JembeWF 'process_model' must be provided in __init__ or in init_app"
            )
        if process_model is not None:
            self.process_model = process_model

        # check if step_model is provided
        if step_model is None and not hasattr(self, "step_model"):
            raise Exception(
                "JembeWF 'step_model' must be provided in __init__ or in init_app"
            )
        if step_model is not None:
            self.step_model = step_model

        # initialise extension
        app.extensions["jembewf"] = self
        self.initialised = True

    def add(self, *flows: "jembewf.Flow") -> "jembewf.JembeWF":
        """Add/Register Flow definition"""
        if self.initialised:
            raise Exception(
                "Can't add flows to JembeWF because it is already initialised."
            )

        for flow in flows:
            if flow.name in self.flows:
                raise Exception(
                    f"Flow with same name '{flow.name}' is already registred."
                )
            self.flows[flow.name] = flow
        return self

    def start(self, flow_name: str, **process_vars) -> "jembewf.ProcessMixin":
        """Start Process instance from Flow definition

        Returns Process instance that has been started.
        """
        return self.process_model.create(flow_name, **process_vars)

    def can_start(self, flow_name: str, **process_vars) -> bool:
        """Check if process from flow definition can be started"""
        return self.process_model.can_start(flow_name, **process_vars)



from typing import TYPE_CHECKING, Optional, Type, List

if TYPE_CHECKING:
    import jembewf

__all__ = ("Task", "TaskCallback")


class TaskCallback:
    """Provides business logic for the Task to which is attached to.

    Extend this class and override methods to provide custom business logic.
    """

    def __init__(self, step: "jembewf.StepMixin"):
        self.step = step
        self.process = step.process

        self.task = step.task
        self.flow = self.task.flow

    def callback(self):
        """Called on arrive to the task"""


class Task:
    """Defines and configures Task"""

    def __init__(
        self,
        name: str,
        callback: Optional[Type["jembewf.TaskCallback"]] = None,
        **config,
    ) -> None:
        self.name = name
        self.callback: Type[TaskCallback] = (
            callback if callback is not None else TaskCallback
        )
        self.config = config
        self.auto_proceed = False

        # list of all transitions that belogns to this task
        self.transitions: List["jembewf.Transition"] = []

        self.flow: "jembewf.Flow"

        self.validate = False

    def add(self, *transitions: "jembewf.Transition") -> "jembewf.Task":
        """Adds transition to the task"""
        self.transitions.extend(transitions)
        return self

    def auto(self) -> "jembewf.Task":
        """Task will proceed automaticaly to its transition when Step is created"""
        self.auto_proceed = True
        return self

    def attach_to_flow(self, flow: "jembewf.Flow"):
        """Attach task to the Flow"""
        self.flow = flow
        self._validate()

    def _validate(self):
        if not hasattr(self, "flow"):
            raise Exception(f"Task '{self.name}' is not attached to the flow")
        self.validate = True



from typing import TYPE_CHECKING, Optional, Type

if TYPE_CHECKING:
    import jembewf

__all__ = ("Transition", "TransitionCallback")


class TransitionCallback:
    """Provides business logic for the Transition to which is attached to.

    Extend this class and override methods to provide custom business logic.
    """

    def __init__(
        self,
        transition: "jembewf.Transition",
        from_step: "jembewf.StepMixin",
    ):
        self.from_step = from_step
        self.process = from_step.process

        self.transition = transition
        self.from_task = transition.from_task
        self.to_task = transition.to_task
        self.flow = transition.flow

        if transition.flow != self.process.flow:
            raise Exception(
                f"Transition flow '{transition.flow.name}'  "
                f"and Process flow '{self.process.flow}' are not equal when "
                f" proceeding from step '{self.from_task.name}' to '{self.to_task.name}'."
            )

    def can_proceed(self) -> bool:
        """Check if a transition is ready proceed to next step/task.

        Default returns True.

        Override this method to implement check to see if all contitions
        for proceeding to next task are meet.

        If you want to autamticaly proceed to next task then override this
        method and return True.
        """
        return True

    def callback(self, to_step: "jembewf.StepMixin"):
        """Called when transiting to the to_task/step"""


class Transition:
    """Defines and configures Transition"""

    def __init__(
        self,
        to_task_name: str,
        callback: Optional[Type["jembewf.TransitionCallback"]] = None,
        **config,
    ) -> None:
        self.to_task_name = to_task_name
        self.callback: Type[TransitionCallback] = (
            callback if callback is not None else TransitionCallback
        )
        self.config = config

        self.flow: "jembewf.Flow"
        self.from_task: "jembewf.Task"
        self.to_task: "jembewf.Task"

        self.validate = False

    def attach_to_from_task(self, task: "jembewf.Task"):
        """Attach to the Task"""
        self.from_task = task
        self.flow = task.flow
        self.to_task = self.flow.tasks[self.to_task_name]
        self._validate()

    def _validate(self):
        if not hasattr(self, "from_task"):
            raise Exception(
                f"Transition '{self.name}' is not attached to the from_task"
            )
        if not hasattr(self, "to_task"):
            raise Exception(
                f"Transition '{self.name}' is not associated with the to_task"
            )
        if not hasattr(self, "flow"):
            raise Exception(f"Transition '{self.name}' is not associated with the flow")
        self.validate = True



from typing import TYPE_CHECKING, Type, Dict, Optional, List

if TYPE_CHECKING:
    import jembewf

__all__ = ("Flow", "FlowCallback")


class FlowCallback:
    """Provides business logic for starting the flow to which is attached to.

    Extend this class and override methods to provide custom business logic
    """

    def __init__(self, process: "jembewf.ProcessMixin"):
        self.process = process

        self.flow = process.flow

    def can_start(self) -> bool:
        """Check if a flow can be started. Default returns True"""
        return True

    def callback(self):
        """Called right after process is created and before startit transitions"""


class Flow:
    """Defines and configures Flow"""

    def __init__(
        self,
        name: str,
        callback: Optional[Type["jembewf.FlowCallback"]] = None,
        **config,
    ) -> None:
        self.name = name
        self.callback: Type[FlowCallback] = callback if callback else FlowCallback
        self.config = config

        # list of all tasks that belongs to this workflow
        self.tasks: Dict[str, "jembewf.Task"] = {}
        self.starts_with_tasks: List[str] = []
        self.ends_with_tasks: List[str] = []

        self.validated = False

    def start_with(self, *task_names: str) -> "jembewf.Flow":
        """Define task names that will be executed when flow starts

        Flow must call start_with in order to have valid and complet definition.
        """
        # at least one task_name
        if len(task_names) == 0:
            raise Exception(
                f"At least one task_name must be provided to flow '{self.name}'."
            )

        # task with provided names exist
        for task_name in task_names:
            if task_name not in self.tasks:
                raise Exception(
                    f"Task with name '{task_name}' doesn't exist in flow '{self.name}'."
                )
        self.starts_with_tasks = list(task_names)


        # Associate from_task, to_task, and flow inside transitions
        # Can't be done earlier because we need to define have
        # all tasks defined to associate to_task 
        for task in self.tasks.values():
            for transition in task.transitions:
                transition.attach_to_from_task(task)

        self._validate_flow()
        return self

    def add(self, *tasks: "jembewf.Task") -> "jembewf.Flow":
        """Add Tasks to flow

        Returns:
            jembewf.Flow: returns self
        """
        for task in tasks:
            if task.name in self.tasks:
                raise Exception(
                    f"Task '{task.name}' already exist in flow '{self.name}'."
                )

            task.attach_to_flow(self)
            self.tasks[task.name] = task
            if len(task.transitions) == 0:
                self.ends_with_tasks.append(task.name)
        return self

    def _validate_flow(self):
        if len(self.tasks) == 0:
            Exception(f"Flow '{self.name} have no tasks.")

        if len(self.starts_with_tasks) == 0:
            Exception(
                f"Flow '{self.name} have no start tasks defined. "
                "Use Flow.start_with to define start tasks."
            )

        if len(self.ends_with_tasks) == 0:
            Exception(
                f"Flow '{self.name} have no endtasks. Tasks are making infinite loop"
            )

        # check if all transitions  lead to to the existing task_name
        for task in self.tasks.values():
            for transition in task.transitions:
                if transition.to_task_name not in self.tasks:
                    raise Exception(
                        f"Transition from task '{task.name}' "
                        f"in flow '{self.name}' leads to non existing "
                        f"task '{transition.to_task_name}'."
                    )
        self.validated = True



from typing import TYPE_CHECKING, Optional
from datetime import datetime
from sqlalchemy_json import NestedMutableJson
from sqlalchemy.orm import declarative_mixin, declared_attr
import sqlalchemy as sa
from .helpers import get_jembewf

if TYPE_CHECKING:
    import jembewf

__all__ = ("StepMixin",)


@declarative_mixin
class StepMixin:
    """Mixin to be applied to Step SqlAlchemy model

    Step model keep track of all instances of tasks in flows (current and completed)
    and it's local variables and states.

    StepMixin should be applied class extended from
    flask_sqlalchemy.SqlAlchemy().Model who defines model in database
    """

    __process_table_name__: str = "jwf_processes"
    __process_class_name__: str = "Process"

    __tablename__ = "jwf_steps"

    id = sa.Column(sa.Integer, primary_key=True)

    # process
    @declared_attr
    def process_id(cls):
        """Foreign key to Process table"""
        return sa.Column(
            sa.Integer,
            sa.ForeignKey(f"{cls.get_process_table_name()}.id"),
            nullable=False,
        )

    @declared_attr
    def process(cls):
        """Defines process relationship"""
        return sa.orm.relationship(
            cls.get_process_class_name(),
            foreign_keys=[cls.process_id],
            back_populates="steps",
        )

    task_name = sa.Column(sa.String(250), nullable=False)

    is_active = sa.Column(
        sa.Boolean(), nullable=False, default=True, server_default=sa.true()
    )
    is_last_step = sa.Column(
        sa.Boolean(), nullable=False, default=False, server_default=sa.false()
    )

    # step/task variables
    variables = sa.Column(NestedMutableJson)

    started_at = sa.Column(sa.DateTime, default=datetime.utcnow, nullable=False)
    ended_at = sa.Column(sa.DateTime)

    # prev_step, next_step

    @declared_attr
    def prev_step_id(cls):
        return sa.Column(
            sa.Integer,
            sa.ForeignKey(f"{cls.__tablename__}.id"),
        )

    @declared_attr
    def prev_step(cls):
        return sa.orm.relationship(
            f"{cls.__name__}", remote_side=[cls.id], foreign_keys=[cls.prev_step_id]
        )

    @property
    def task(self) -> "jembewf.Task":
        """Returns Task definition of the step"""
        return self.process.flow.tasks[self.task_name]

    @property
    def callback(self) -> "jembewf.TaskCallback":
        """TaskCallback of the task instance"""
        return self.task.callback(self)

    @classmethod
    def create(
        cls,
        process: "jembewf.ProcessMixin",
        task: "jembewf.Task",
        prev_step: Optional["jembewf.StepMixin"] = None,
        transition_callback: Optional["jembewf.TransitionCallback"] = None,
        **step_vars,
    ) -> "jembewf.StepMixin":
        """Creates step instance in process for the task

        Args:
            process (jembewf.ProcessMixin): Process instance to whome the step will belong
            task (jembewf.Task): Task instance for wichin we create the step
            from_step (Optional[jembewf.StepMixin]): previous step

        Returns:
            jembewf.StepMixin: _description_
        """
        jwf = get_jembewf()
        step = jwf.step_model()
        step.task_name = task.name
        step.process = process
        step.variables = step_vars
        if prev_step:
            step.prev_step = prev_step
        step.is_last_step = len(task.transitions) == 0

        jwf.db.session.add(step)
        jwf.db.session.commit()

        if transition_callback:
            transition_callback.callback(step)

        step.callback.callback()

        if step.is_last_step:
            step.is_active = False
            step.ended_at = datetime.utcnow()
            jwf.db.session.add(step)
            jwf.db.session.commit()
        elif step.task.auto_proceed:
            step.proceed()

        return step

    def proceed(self) -> bool:
        """proceed with the process execution by transiting from this step forward

        Returns True if we proceeded to next steps
        """
        # check if this is not last step and is active
        if self.is_last_step or not self.is_active:
            return False

        proceeded = False
        for transition in self.task.transitions:
            transition_callback = transition.callback(transition, self)
            if transition_callback.can_proceed():
                self.create(self.process, transition.to_task, self, transition_callback)
                proceeded = True

        if proceeded:
            self.is_active = False
            self.ended_at = datetime.utcnow()
            jwf = get_jembewf()
            jwf.db.session.add(self)
            jwf.db.session.commit()
            self.process.check_is_running()

        return proceeded

    @classmethod
    def get_process_table_name(cls) -> str:
        """Returns name of Process table defined in cls.__process_table_name__

        It's used to create relationship between steps and processes.
        """
        try:
            return cls.__process_table_name__
        except AttributeError as err:
            raise AttributeError(
                f"Attribute __process_table_name__ for '{cls.__name__}' is not defined"
            ) from err

    @classmethod
    def get_process_class_name(cls) -> str:
        """Returns name of Process class defined in cls.__process_class_name__

        It's used to create relationship between steps and processes.
        """
        try:
            return cls.__process_class_name__
        except AttributeError as err:
            raise AttributeError(
                f"Attribute __process_class_name__ for '{cls.__name__}' is not defined"
            ) from err

    def __repr__(self):
        return f"<Step #{self.id}: '{self.task_name}' from process #{self.process_id}: '{self.process.flow_name}'>"



from typing import TYPE_CHECKING
from flask import current_app

if TYPE_CHECKING:
    import jembewf

__all__ = ("get_jembewf",)


def get_jembewf() -> "jembewf.JembeWF":
    """Returns instance of JembeWf for current Flask application"""
    jembewf_instance = current_app.extensions.get("jembewf", None)
    if jembewf_instance is None:
        raise Exception("JembeWF extension is not initialised")
    return jembewf_instance
